/*
 * hexinject.c
 *
 *  Created on: 11/mag/2010
 *      Author: Acri Emanuele
 */

#include "hexinject.h"
#include "hexstring.h"
#include "hexdump.h"
#include "argparser.h"
#include "prettypacket.h"

/*
 * Checksum IP
 */
uint16_t ip_cksum (uint16_t *buff, size_t len) {
    
    uint32_t sum = 0;
    uint16_t answer = 0;

    while(len > 1) {
        sum += *buff++;
        len -= 2;
    }

    if (len) {
        sum += * (uint8_t *) buff;
    }

    while (sum>>16)
        sum = (sum & 0xffff) + (sum >> 16);

    answer = ~sum;

    return(answer);
}

/*
 * Checksum TCP
 */
uint16_t tcp_cksum(uint16_t *src_addr, uint16_t *dest_addr, uint16_t *buff, uint16_t len) {

    uint32_t sum = 0;
    uint16_t answer = 0;

    sum += src_addr[0];
    sum += src_addr[1];
    
    sum += dest_addr[0];
    sum += dest_addr[1];

    sum += htons(0x6);

    sum += htons(len);

    while(len > 1) {
        sum += *buff++;
        len -= 2;
    }

    if (len) {
        sum += * (uint8_t *) buff;
    }

    while (sum>>16)
        sum = (sum & 0xffff) + (sum >> 16);

    answer = ~sum;

    return(answer);
}

/*
 * Checksum UDP
 */
uint16_t udp_cksum(uint16_t *src_addr, uint16_t *dest_addr, uint16_t *buff, size_t len) {
   
    uint32_t sum = 0;
    uint16_t answer = 0;
    
    sum += src_addr[0];
    sum += src_addr[1];
    
    sum += dest_addr[0];
    sum += dest_addr[1];

    sum += htons(0x11);

    sum += htons(len);

    while(len > 1) {
        sum += *buff++;
        len -= 2;
    }

    if (len) {
        sum += * (uint8_t *) buff;
    }

    while (sum>>16)
        sum = (sum & 0xffff) + (sum >> 16);

    answer = ~sum;

    return(answer);
}

/*
 * Do checksum (if the packet requires it...)
 */
void do_cksum (char *raw, size_t size) {
    
    uint16_t *cksum = NULL;

    // disabled?
    if (options.no_cksum) {
        return;
    }

    // is ip?
    if ( size >= 34 && raw[12]==0x08  && raw[13]==0x00  ) {
       
        // ip checksum
        cksum = (uint16_t *) &raw[24];
        *cksum = 0;

        *cksum = ip_cksum((uint16_t *) &raw[14], 20);

        // next protocol
        switch(raw[23]) {

            // tcp
            case 0x06:
                if (size < 54) return; // size check
                cksum = (uint16_t *) &raw[50];
                *cksum = 0;
                *cksum = tcp_cksum((uint16_t *) &raw[26], (uint16_t *) &raw[30], (uint16_t *) &raw[34], (size-34));
                break;

            // udp
            case 0x11:
                if (size < 42) return; // size check
                cksum = (uint16_t *) &raw[40];
                *cksum = 0;
                *cksum = udp_cksum((uint16_t *) &raw[26], (uint16_t *) &raw[30], (uint16_t *) &raw[34], (size-34));
                break;

            // icmp
            case 0x01:
                if (size < 42) return; // size check
                cksum = (uint16_t *) &raw[36];
                *cksum = 0;
                *cksum = ip_cksum((uint16_t *) &raw[34], (size-34));
                break;
        }
    }
}

/*
 * Adjust packet size fields (if the packet requires it...)
 */
void do_size (char *raw, size_t size) {
    
    uint16_t *len_field = NULL;

    // disabled?
    if (options.no_size) {
        return;
    }

    // is ip?
    if ( size >= 34 && raw[12]==0x08  && raw[13]==0x00  ) {
       
        // ip total length
        len_field = (uint16_t *) &raw[16];

        *len_field = size - 14; // size - ethernet header
        *len_field = htons(*len_field);

        // next protocol
        switch(raw[23]) {

            // tcp
            case 0x06:
                if (size < 54) return; // size check
                // tcp uses header length field
                break;

            // udp
            case 0x11:
                if (size < 42) return; // size check
                len_field = (uint16_t *) &raw[38];
                *len_field = size - 14 - ((raw[14] & 0xF) * 4); // size - ethernet header - ip header
                *len_field = htons(*len_field);
                break;

            // icmp
            case 0x01:
                if (size < 42) return; // size check
                // no size field
                break;
        }
    }
}

/*
 * Inject a raw buffer to the network
 */
int inject_raw(pcap_t *fp, char *raw, size_t size) {

	assert(fp != NULL);
	assert(raw != NULL);

	int err = 0;
    
    /* packet size (if enabled) */
    do_size (raw, size);

    /* checksum */
    do_cksum (raw, size);

	/* Send down the packet */
	err = pcap_sendpacket(fp, (unsigned char *) raw, size);

	return err;
}

/*
 * Inject an hexstring to the network
 */
int inject_hexstr(pcap_t *fp, char *hexstr) {

	assert(fp != NULL);
	assert(hexstr != NULL);

	int err = 0;
	int size = 0;
	char *raw = NULL;

	raw = hexstr_to_raw(hexstr, &size);

	/* Send down the packet */
	err = inject_raw(fp, raw, size);

	free(raw);

	return err;
}

/*
 * Sniff a packet from the network. Hexstring mode.
 */
char *sniff_hexstr(pcap_t *fp) {

	assert(fp != NULL);

	struct pcap_pkthdr hdr;
	char *hexstr = NULL;
	char *raw    = NULL;

	/* Sniff the packet */
	raw = (char *) pcap_next(fp, &hdr);

    if(raw == NULL)
    	return NULL;

    hexstr = raw_to_hexstr(raw, hdr.len);

    return hexstr;
}

/*
 * Sniff a packet from the network. Raw mode.
 */
char *sniff_raw(pcap_t *fp, size_t *size) {

	assert(fp != NULL);
	assert(size != NULL);

	struct pcap_pkthdr hdr;
	char *raw = NULL;

	/* Sniff the packet */
	raw = (char *) pcap_next(fp, &hdr);

	*size = hdr.len;

    return raw;
}

/*
 * Convert standard input in hexstring format to raw format
 */
int convert_to_raw_loop()
{
    char buffer[BUFFER_SIZE];
    
    while (!feof(stdin)) {
    
        /* Read hexstring */
        if( !fgets(buffer, BUFFER_SIZE, stdin) ) {
            continue;
        }

        /* Write raw output */
    	int size = 0;
    	char *raw = NULL;

    	raw = hexstr_to_raw(buffer, &size);

    	fwrite(raw, 1, size, stdout);
        fflush(stdout);

    	free(raw);
    }

    return 0;
}

/*
 * Convert standard input in raw format to hexstring format
 */
int convert_to_hexstr_loop()
{
    char buffer[BUFFER_SIZE];
    int size;
    
    while (!feof(stdin)) {
        
        /* Read raw */
        size = fread(buffer, 1, BUFFER_SIZE, stdin);

        /* Write hexstring output */
        char *hexstr = raw_to_hexstr(buffer, size);
        
        fprintf(stdout, "%s\n", hexstr);
        fflush(stdout);

        free(hexstr);
    }

    return 0;
}

/*
 * Injection loop using hexstring format
 */
int inject_hexstr_loop(pcap_t *fp)
{
    char buffer[BUFFER_SIZE];

    while (!feof(stdin)) {

        /* Read hexstring */
        if( !fgets(buffer, BUFFER_SIZE, stdin) ) {
            continue;
        }

        /* Send down the packet */
        if (inject_hexstr(fp, buffer) != 0) {
            fprintf(stderr,"\nError sending the packet: %s\n", pcap_geterr(fp));
            return 1;
        }
    }

    return 0;
}

/*
 * Injection loop using raw format
 */
int inject_raw_loop(pcap_t *fp)
{
    char buffer[BUFFER_SIZE];
    size_t size;

    while (!feof(stdin)) {

        /* Read raw */
        size = fread(buffer, 1, BUFFER_SIZE, stdin);

        /* Send down the packet */
        if (inject_raw(fp, buffer, size) != 0) {
            fprintf(stderr,"\nError sending the packet: %s\n", pcap_geterr(fp));
            return 1;
        }
    }

    return 0;
}

/*
 * Sniffing loop with hexstring output
 */
int sniff_hexstr_loop(pcap_t *fp)
{
    char *hexstr;
    
    while ( 1 ) {

        /* Count */
        if (options.count_on) {
            if (options.count <= 0) {
                break;
            }
        }

        /* Sniff and print a packet */
        if ((hexstr = sniff_hexstr(fp))) {
            printf("%s\n", hexstr);
            fflush(stdout);
            free(hexstr);

            /* Count */
            if (options.count_on) {
                options.count--;
            }
        }

        usleep(options.sleep_time);
    }

    return 0;
}

/*
 * Sniffing loop with raw output
 */
int sniff_raw_loop(pcap_t *fp)
{
    char *packet;
    size_t size;

    while ( 1 ) {

        /* Count */
        if (options.count_on) {
            if (options.count <= 0) {
                break;
            }
        }

        size = BUFFER_SIZE;

        /* Sniff and print a packet */
        if ((packet = sniff_raw(fp, &size))) {
            fwrite(packet, 1, size, stdout);
            fflush(stdout);

            /* Count */
            if (options.count_on) {
                options.count--;
            }
        }

        usleep(options.sleep_time);
    }

    return 0;
}

/*
 * Sniffing loop with raw outputi,
 * plus pretty printing of disassembled packet
 */
int sniff_raw_dis_loop(pcap_t *fp)
{
    char *packet;
    size_t size;

    while ( 1 ) {

        /* Count */
        if (options.count_on) {
            if (options.count <= 0) {
                break;
            }
        }

        size = BUFFER_SIZE;

        /* Sniff and print a packet */
        if ((packet = sniff_raw(fp, &size))) {
            ethernet_print(packet, size);
            
            /* Print end of packet */
            puts("\n ----------- ");
            fflush(stdout);

            /* Count */
            if (options.count_on) {
                options.count--;
            }
        }

        usleep(options.sleep_time);
    }

    return 0;
}

/*
 * Print disassembled example packet
 */
void print_dis_example_packet (enum packet_type type) {
    switch (type) {
        case tcp:  ethernet_print(tcp_packet, sizeof(tcp_packet)-1); break;
        case udp:  ethernet_print(udp_packet, sizeof(icmp_packet)-1); break;
        case icmp: ethernet_print(icmp_packet, sizeof(udp_packet)-1); break;
        case arp:  ethernet_print(arp_packet, sizeof(arp_packet)-1); break;
    }
    puts("");
}

/*
 * Main function
 */
int main(int argc, char **argv) {

    pcap_t *fp;
    struct bpf_program bpf;
    char errbuf[PCAP_ERRBUF_SIZE];

    int ret_val;

    char *dev;
	
    /* Parse cmdline options */
    parseopt(argc, argv);
    
    /* in case of conversion mode */
    if(options.convert_to_raw) {
        return convert_to_raw_loop();
    }
    else if(options.convert_to_hexstr) {
        return convert_to_hexstr_loop();
    }

    /* in case of prettyprinting a single disassembled packet  of example */
    if(options.print_dis_example != no_type) {
        print_dis_example_packet(options.print_dis_example);
        return 0;
    }
    
    /* find a device if not specified */
    if(!options.device) {
        dev = pcap_lookupdev(errbuf);
        if (dev == NULL) {
            fprintf(stderr,"Unable to find a network adapter: %s.\n", errbuf);
            return 1;
        }
    }
    else {
        dev = options.device;
    }

    /* Create packet capture handle */
    if((fp = pcap_create(dev, errbuf)) == NULL) {
        fprintf(stderr,"Unable to create pcap handle: %s\n", errbuf);
        return 1;
    }

    /* Set snapshot length */
    if(pcap_set_snaplen(fp, BUFSIZ) != 0) {
        fprintf(stderr,"Unable to set snapshot length: the interface may be already activated\n");
        return 1;
    }

    /* Set promiscuous mode */
    if(pcap_set_promisc(fp, options.promisc) != 0) {
        fprintf(stderr,"Unable to set promiscuous mode: the interface may be already activated\n");
        return 1;
    }

    /* Set read timeout */
    if(pcap_set_timeout(fp, 1000) != 0) { // a little patch i've seen in freebsd ports: thank you guys ;)
        fprintf(stderr,"Unable to set read timeout: the interface may be already activated\n");
        return 1;
    }

    /* Set monitor mode */
    if(options.monitor) {
        if(pcap_can_set_rfmon(fp)==0) {
            fprintf(stderr, "Monitor mode not supported by %s.\n", dev);
            return 1;
        }

        if((ret_val=pcap_set_rfmon(fp, 1)) != 0) {
            fprintf(stderr, "Unable to set monitor mode: the interface may be already activated.\n");
            return 1;
        }
    }

    /* Activate interface */
    if(pcap_activate(fp) != 0) {
        fprintf(stderr, "Unable to activate the interface: %s\n", pcap_geterr(fp));
        return 1;
    }
    
    /* Apply filter */
    if ( options.filter ) {

        if(pcap_compile(fp, &bpf, options.filter, 0, 0) != 0) {
            fprintf(stderr, "Error compiling filter: %s\n", pcap_geterr(fp));
            return 1;
        }

        if(pcap_setfilter(fp, &bpf) != 0) {
            fprintf(stderr, "Error setting filter: %s\n", pcap_geterr(fp));
            return 1;
        }

    }

    /* Inject mode - Hexstring */
    if (options.inject && !options.raw) {
        ret_val = inject_hexstr_loop(fp);
    }
    
    /* Inject mode - Raw */
    else if (options.inject && options.raw) {
        ret_val = inject_raw_loop(fp);
    }
    
    /* Sniff mode - Hexstring */
    else if (options.sniff && !options.raw && !options.print_dis) {
        ret_val = sniff_hexstr_loop(fp);    
    }
    
    /* Sniff mode - Raw */
    else if (options.sniff && options.raw && !options.print_dis) {
        ret_val = sniff_raw_loop(fp);
    }

    /* Sniff mode - pretty printing, disassembling */
    else if (options.sniff && options.print_dis) {
        ret_val = sniff_raw_dis_loop(fp);
    }

    pcap_close(fp);

    return ret_val;
}
